from distutils.core import setup

setup(
  name='pyDPMP',
  version='0.1.2',
  description='The Diverse Particle Max-Product algorithm (D-PMP)',
  author='Samuel Ainsworth',
  author_email='skainsworth@gmail.com',
  url='https://github.com/samuela/pyDPMP',
  packages=['pyDPMP'],
  install_requires=['numpy']
)
